import mock

from incuna_auth.middleware import permission
from incuna_auth.middleware.utils import compile_urls
from .utils import RequestTestCase


class TestBasePermissionMiddleware(RequestTestCase):
    middleware_class = permission.BasePermissionMiddleware
    middleware_path = 'incuna_auth.middleware.permission.BasePermissionMiddleware.{}'

    def setUp(self):
        self.middleware = self.middleware_class()

    def test_deny_access(self):
        """Assert that a typical (GET) redirect defaults to /?next=/."""
        response = self.middleware.deny_access(self.create_request())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['location'], '/?next=/')

    def test_deny_access_post(self):
        """Assert that a POST request results in a HTTP 403 (Forbidden) error."""
        response = self.middleware.deny_access(self.create_request(method='post'))
        self.assertEqual(response.status_code, 403)

    def test_deny_access_no_message(self):
        """Assert that a blank `message` parameter results in no Django message."""
        request = self.create_request()
        self.middleware.deny_access(request)
        self.assertEqual(request._messages.store, [])

    def test_deny_access_message(self):
        """Assert that a non-blank `message` parameter results in a Django message."""
        request = self.create_request()
        method = self.middleware_path.format('get_access_denied_message')
        message = 'I am a message'
        with mock.patch(method, return_value=message):
            self.middleware.deny_access(request)
            self.assertEqual(request._messages.store, [message])

    def test_process_request_default(self):
        """Assert that the default process_request implementation does nothing."""
        response = self.middleware.process_request(self.create_request())
        self.assertIsNone(response)

    def test_process_request_default_protected(self):
        """
        Assert that the default process_request implementation does nothing even if the
        resource is protected.
        """
        method = self.middleware_path.format('is_resource_protected')
        with mock.patch(method, return_value=True):
            response = self.middleware.process_request(self.create_request())
            self.assertIsNone(response)

    def test_process_request_denies_access(self):
        """
        Assert that the default process_request implementation denies access when both
        is_resource_protected and deny_access_condition return True.
        """
        resource_method = self.middleware_path.format('is_resource_protected')
        condition_method = self.middleware_path.format('deny_access_condition')
        with mock.patch(resource_method, return_value=True):
            with mock.patch(condition_method, return_value=True):
                response = self.middleware.process_request(self.create_request())
                self.assertEqual(response.status_code, 302)

    def test_default_unauthorised_redirect_url(self):
        """
        Assert the default implementation of get_unauthorised_redirect_url returns '/'.
        """
        request = self.create_request()
        self.assertEqual('/', self.middleware.get_unauthorised_redirect_url(request))

    def test_configured_unauthorised_redirect_url(self):
        """
        Assert get_unauthorised_redirect_url returns base_unauthorised_redirect_url.
        """
        request = self.create_request()
        url = '/redirect/'
        self.middleware.base_unauthorised_redirect_url = url
        self.assertEqual(url, self.middleware.get_unauthorised_redirect_url(request))

    def test_access_denied_message(self):
        """
        Assert the default implementation of get_access_denied_message returns ''.
        """
        self.assertEqual('', self.middleware.get_access_denied_message(request=None))


class TestLoginMiddlewareMixin(RequestTestCase):
    middleware_class = permission.LoginPermissionMiddlewareMixin

    def setUp(self):
        # We don't need to mix this into anything because we're only calling the methods
        # it implements.
        self.middleware = self.middleware_class()

    def test_deny_access_condition(self):
        """Assert that deny_access_condition disallows an anonymous user."""
        request = self.create_request(auth=False)
        self.assertTrue(self.middleware.deny_access_condition(request))

    def test_deny_access_condition_allow(self):
        """Assert that deny_access_condition allows an authenticated user."""
        request = self.create_request(auth=True)
        self.assertFalse(self.middleware.deny_access_condition(request))

    def test_access_denied_message(self):
        """Assert the message returned by get_access_denied_message."""
        expected_message = 'You must be logged in to view this page.'
        self.assertEqual(
            expected_message,
            self.middleware.get_access_denied_message(request=None)
        )


class TestUrlPermissionMiddleware(RequestTestCase):
    middleware_class = permission.UrlPermissionMiddleware
    middleware_path = 'incuna_auth.middleware.permission.UrlPermissionMiddleware.{}'
    URL = '/fake-request/'
    ALL_URLS = compile_urls([r'^'])

    def setUp(self):
        self.middleware = self.middleware_class()

    def make_request(self, url=URL, **kwargs):
        """Create a request with a suitable URL attribute."""
        request = self.create_request(url=url, **kwargs)
        return request

    def test_url_exempt(self):
        """
        Assert that the URL is not protected when the request matches an item in
        exempt_urls.
        """
        request = self.make_request()
        method = self.middleware_path.format('get_exempt_url_patterns')
        with mock.patch(method, return_value=self.ALL_URLS):
            self.assertFalse(self.middleware.is_resource_protected(request))

    def test_url_protected(self):
        """
        Assert that the URL is protected when the request matches an item in
        protected_urls.
        """
        request = self.make_request()
        self.assertTrue(self.middleware.is_resource_protected(request))

    def test_url_unprotected(self):
        """
        Assert that the URL is not protected when the request matches neither exempt_urls
        nor protected_urls.
        """
        request = self.make_request()
        method = self.middleware_path.format('get_protected_url_patterns')
        with mock.patch(method, return_value=[]):
            self.assertFalse(self.middleware.is_resource_protected(request))



import re
from base64 import b64encode
from unittest import TestCase

import mock
from django.conf import settings
from django.test.utils import override_settings

from incuna_auth.middleware import (
    basic_auth,
    FeinCMSLoginRequiredMiddleware,
    LoginRequiredMiddleware,
)
from incuna_auth.models import AccessStateExtensionMixin as AccessState
from .utils import RequestTestCase

NO_URLS = []
ALL_URLS = [re.compile(r'^')]


def base64_encode_for_py2or3(text):
    """Encode a base-64 string in a way that works in Python 2 or 3."""
    return b64encode(text.encode('utf-8')).decode('utf-8')


class TestLoginRequiredMiddleware(RequestTestCase):
    middleware = LoginRequiredMiddleware(check=False)
    EXEMPT_URLS = 'incuna_auth.middleware.LoginRequiredMiddleware.EXEMPT_URLS'
    PROTECTED_URLS = 'incuna_auth.middleware.LoginRequiredMiddleware.PROTECTED_URLS'

    def make_request(self, auth, method='get', url='/fake-request/', **kwargs):
        return self.create_request(method, auth=auth, url=url, **kwargs)

    @mock.patch(EXEMPT_URLS, ALL_URLS)
    def test_exempt_url(self):
        request = self.make_request(auth=False)
        response = self.middleware.process_request(request)
        self.assertIsNone(response)

    @mock.patch(EXEMPT_URLS, NO_URLS)
    @mock.patch(PROTECTED_URLS, NO_URLS)
    def test_unprotected_url(self):
        request = self.make_request(auth=False)
        response = self.middleware.process_request(request)
        self.assertIsNone(response)

    @mock.patch(EXEMPT_URLS, NO_URLS)
    @mock.patch(PROTECTED_URLS, ALL_URLS)
    def test_is_auth(self):
        request = self.make_request(auth=True)
        response = self.middleware.process_request(request)
        self.assertIsNone(response)

    @mock.patch(EXEMPT_URLS, NO_URLS)
    @mock.patch(PROTECTED_URLS, ALL_URLS)
    def test_non_auth_get(self):
        request = self.make_request(auth=False)
        response = self.middleware.process_request(request)
        message = 'You must be logged in to view this page.'
        redirect_url = settings.LOGIN_URL + '?next=/fake-request/'

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['location'], redirect_url)
        self.assertEqual(request._messages.store, [message])

    @mock.patch(EXEMPT_URLS, NO_URLS)
    @mock.patch(PROTECTED_URLS, ALL_URLS)
    def test_non_auth_post(self):
        request = self.make_request(auth=False, method='post')
        response = self.middleware.process_request(request)
        self.assertEqual(response.status_code, 403)


class TestFeinCMSLoginRequiredMiddleware(RequestTestCase):
    middleware = FeinCMSLoginRequiredMiddleware()
    AUTH_STATE = AccessState.STATE_AUTH_ONLY
    OTHER_STATE = ('other', 'Other state')

    get_page_method = (
        'incuna_auth.middleware.permission_feincms.' +
        'FeinCMSPermissionMiddleware._get_page_from_path'
    )

    class DummyFeinCMSPage:
        def __init__(self, access_state):
            self.access_state = access_state

    def make_request(self, auth, access_state, **kwargs):
        """Create a request with a suitable feincms_page attribute."""
        request = self.create_request(auth=auth, **kwargs)
        request.feincms_page = self.DummyFeinCMSPage(access_state=access_state)
        return request

    def test_unprotected_url(self):
        request = self.make_request(
            auth=False,
            access_state=AccessState.STATE_ALL_ALLOWED
        )
        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            response = self.middleware.process_request(request)
        self.assertIsNone(response)

    def test_other_state_url(self):
        request = self.make_request(auth=False, access_state=self.OTHER_STATE)
        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            response = self.middleware.process_request(request)
        self.assertIsNone(response)

    def test_is_auth(self):
        request = self.make_request(auth=True, access_state=self.AUTH_STATE)
        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            response = self.middleware.process_request(request)
        self.assertIsNone(response)

    def test_non_auth_get(self):
        request = self.make_request(auth=False, access_state=self.AUTH_STATE)
        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            response = self.middleware.process_request(request)
        message = 'You must be logged in to view this page.'
        redirect_url = settings.LOGIN_URL + '?next=/'

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['location'], redirect_url)
        self.assertEqual(request._messages.store, [message])

    def test_non_auth_post(self):
        request = self.make_request(
            auth=False,
            access_state=self.AUTH_STATE,
            method='post'
        )
        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            response = self.middleware.process_request(request)
        self.assertEqual(response.status_code, 403)


class TestBasicAuthMiddleware(TestCase):

    class DummyRequest(object):
        META = {}

        def __init__(self, auth_method=None, username='', password=''):
            if auth_method is None:
                return

            auth = username + ':' + password
            http_auth_string = auth_method + ' ' + base64_encode_for_py2or3(auth)
            self.META = {'HTTP_AUTHORIZATION': http_auth_string}

    def setUp(self):
        self.middleware = basic_auth.BasicAuthenticationMiddleware()

    def test_challenge(self):
        response = basic_auth.challenge()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(response['WWW-Authenticate'], 'Basic realm="Restricted Access"')

    def test_is_authenticated_success(self):
        request = self.DummyRequest('basic', 'user', 'pass')
        result = basic_auth.is_authenticated(request.META['HTTP_AUTHORIZATION'])
        self.assertTrue(result)

    def test_authenticate_failure_wrong_type(self):
        request = self.DummyRequest('non_basic', 'user', 'pass')
        result = basic_auth.is_authenticated(request.META['HTTP_AUTHORIZATION'])
        self.assertEqual(result, None)

    def test_is_authenticated_failure_wrong_credentials(self):
        request = self.DummyRequest('basic', 'other_user', 'other_pass')
        result = basic_auth.is_authenticated(request.META['HTTP_AUTHORIZATION'])
        self.assertFalse(result)

    @override_settings(BASIC_WWW_AUTHENTICATION=False)
    def test_lack_of_www_authentication(self):
        request = self.DummyRequest('basic', 'other_user', 'other_pass')
        result = self.middleware.process_request(request)
        self.assertEqual(result, None)

    def test_no_http_auth_in_meta(self):
        request = self.DummyRequest(None)
        result = self.middleware.process_request(request)
        self.assertEqual(result.status_code, basic_auth.challenge().status_code)

    def test_passes_authentication(self):
        request = self.DummyRequest('basic', 'user', 'pass')
        result = self.middleware.process_request(request)
        self.assertEqual(result, None)

    def test_falls_through_to_challenge(self):
        request = self.DummyRequest('basic', 'other_user', 'other_pass')
        result = self.middleware.process_request(request)
        self.assertEqual(result.status_code, basic_auth.challenge().status_code)



import sys

import dj_database_url
import django
from colour_runner.django_runner import ColourRunnerMixin
from django.conf import settings


settings.configure(
    # Core environmental settings
    DATABASES={
        'default': dj_database_url.config(default='postgres://localhost/incuna_auth'),
    },
    INSTALLED_APPS=(
        'incuna_auth',

        'incuna_test_utils',

        # Put contenttypes before auth to work around test issue.
        # See: https://code.djangoproject.com/ticket/10827#comment:12
        'django.contrib.contenttypes',
        'django.contrib.auth',
        'django.contrib.sessions',
        'django.contrib.admin',
    ),
    PASSWORD_HASHERS=('django.contrib.auth.hashers.MD5PasswordHasher',),
    ROOT_URLCONF='incuna_auth.urls',
    MIDDLEWARE_CLASSES=(),
    REST_FRAMEWORK={
        'DEFAULT_PERMISSION_CLASSES': ('rest_framework.permissions.IsAuthenticated',),
        'DEFAULT_RENDERER_CLASSES': ('rest_framework.renderers.JSONRenderer',),
    },

    # LoginRequiredMiddleware data
    LOGIN_REQUIRED_SEND_MESSAGE=False,

    # BasicAuthenticationMiddleware data
    BASIC_WWW_AUTHENTICATION_USERNAME='user',
    BASIC_WWW_AUTHENTICATION_PASSWORD='pass',
    BASIC_WWW_AUTHENTICATION=True,
)


if django.VERSION >= (1, 7):
    django.setup()

try:
    from django.test.runner import DiscoverRunner
except ImportError:
    # Django < 1.6
    from discover_runner import DiscoverRunner


class Runner(ColourRunnerMixin, DiscoverRunner):
    pass

test_runner = Runner(verbosity=1)
failures = test_runner.run_tests(['tests'])
if failures:
    sys.exit(1)



from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from django.test.utils import override_settings

from incuna_auth.middleware import LoginRequiredMiddleware


class TestCheckRequestHasUser(TestCase):
    middleware = LoginRequiredMiddleware
    requirement = 'django.contrib.auth.middleware.AuthenticationMiddleware'

    @override_settings(MIDDLEWARE_CLASSES=[requirement])
    def test_check_passes(self):
        """Assert that no error is thrown by __init__."""
        self.middleware()

    @override_settings(MIDDLEWARE_CLASSES=[])
    def test_check_fails(self):
        expected_error = ' '.join((
            "MIDDLEWARE_CLASSES does not contain AuthenticationMiddleware.",
            "LoginRequiredMiddleware requires authentication middleware to be",
            "installed. Ensure that your MIDDLEWARE_CLASSES setting includes",
            "'django.contrib.auth.middleware.AuthenticationMiddleware'.",
        ))
        with self.assertRaisesRegexp(ImproperlyConfigured, expected_error):
            self.middleware()



from .permission import BasePermissionMiddleware
from ..models import AccessStateExtensionMixin as AccessState


class FeinCMSPermissionMiddleware(BasePermissionMiddleware):
    """
    Middleware that allows or denies access based on the resource's access state.

    Access state is provided by incuna_auth.models.AccessStateExtensionMixin and is used
    as a marker to determine how a FeinCMS resource should be access-controlled.

    The reasoning behind this is twofold. FeinCMS resources have unpredictable URLs, so
    protecting them with the URL-based middleware is risky. Additionally, a client might
    want to add a variety of permissions on their Pages (or similar) that don't
    necessarily correspond to their location within the site. Adding an access_state
    and checking that through middleware allows maximum flexibility and puts permissions
    on the page level fully within the user's control.

    This class protects all pages with an access_state of STATE_AUTH_ONLY. To protect
    a different state or list of states, override get_protected_states.
    """
    def get_protected_states(self):
        """
        Returns a list of access states this middleware should apply to.

        By default, returns STATE_AUTH_ONLY, which is the only non-custom access state
        that implies any restriction.
        """
        return [AccessState.STATE_AUTH_ONLY]

    def _get_page_from_path(self, path):
        """
        Fetches the FeinCMS Page object that the path points to.

        Override this to deal with different types of object from Page.
        """
        from feincms.module.page.models import Page
        try:
            return Page.objects.best_match_for_path(path)
        except Page.DoesNotExist:
            return None

    def _get_resource_access_state(self, request):
        """
        Returns the FeinCMS resource's access_state, following any INHERITed values.

        Will return None if the resource has an access state that should never be
        protected. It should not be possible to protect a resource with an access_state
        of STATE_ALL_ALLOWED, or an access_state of STATE_INHERIT and no parent.

        Will also return None if the accessed URL doesn't contain a Page.
        """
        feincms_page = self._get_page_from_path(request.path_info.lstrip('/'))
        if not feincms_page:
            return None

        # Chase inherited values up the tree of inheritance.
        INHERIT = AccessState.STATE_INHERIT
        while feincms_page.access_state == INHERIT and feincms_page.parent:
            feincms_page = feincms_page.parent

        # Resources with STATE_ALL_ALLOWED or STATE_INHERIT and no parent should never be
        # access-restricted. This code is here rather than in is_resource_protected to
        # emphasise its importance and help avoid accidentally overriding it.
        never_restricted = (INHERIT, AccessState.STATE_ALL_ALLOWED)
        if feincms_page.access_state in never_restricted:
            return None

        # Return the found value.
        return feincms_page.access_state

    def is_resource_protected(self, request, **kwargs):
        """
        Determines if a resource should be protected.

        Returns true if and only if the resource's access_state matches an entry in
        the return value of get_protected_states().
        """
        access_state = self._get_resource_access_state(request)
        protected_states = self.get_protected_states()
        return access_state in protected_states



from .basic_auth import BasicAuthenticationMiddleware
from .login_required import LoginRequiredMiddleware
from .login_required_feincms import FeinCMSLoginRequiredMiddleware


__all__ = [
    'BasicAuthenticationMiddleware',
    'LoginRequiredMiddleware',
    'FeinCMSLoginRequiredMiddleware',
]



from unittest import TestCase

from incuna_auth.backends import CustomUserModelBackend

from .factories import UserFactory


class TestCustomUserModelBackend(TestCase):
    backend = CustomUserModelBackend()

    def test_nonexistent_user(self):
        result = self.backend.authenticate('fake-user', 'fake-password')
        self.assertEqual(result, None)

    def test_user_by_username(self):
        username = 'user'
        password = 'pass'
        user = UserFactory.create(username=username)
        user.set_password(password)
        user.save()
        result = self.backend.authenticate(username, password)
        self.assertEqual(result, user)

    def test_user_by_email(self):
        email = 'user2@name.com'
        password = 'pass'
        user = UserFactory.create(email=email)
        user.set_password(password)
        user.save()
        result = self.backend.authenticate(email, password)
        self.assertEqual(result, user)

    def test_user_wrong_password(self):
        username = 'user3'
        password = 'pass'
        user = UserFactory.create(username=username)
        user.set_password(password)
        user.save()
        result = self.backend.authenticate(username, 'wrong_password')
        self.assertEqual(result, None)



from django.contrib import messages
from django.contrib.auth.views import redirect_to_login
from django.http import HttpResponseForbidden
from django.utils.translation import ugettext_lazy as _

from .utils import compile_urls


ALL_URLS = compile_urls([r'^'])


class BasePermissionMiddleware:
    """
    Base class for middleware that allows or denies access to a resource.

    This is a generic, fairly useless base class. It protects no pages and allows all
    users through, although if by some miracle it does deny access to a user, it'll be
    fine doing that.

    Contains the following implemented methods:
    - process_request: the method called on incoming request.
    - deny_access: provides standard "you're not allowed" responses.

    Contains the following hook methods:
    - is_resource_protected: Returns True if and only if the middleware should be
      protecting the resource the request is trying to access.
    - deny_access_condition: Returns True if and only if the request should be disallowed
      from accessing the protected resource.
    - get_unauthorised_redirect_url: Returns the URL to redirect a denied GET request to
      (default implementation returns self.base_unauthorised_redirect_url).
    - get_access_denied_message: Returns the message to display when that happens
      (default implementation returns '').

    Contains the following attributes:
    - base_unauthorised_redirect_url: the URL to redirect a denied GET request to if
      get_unauthorised_redirect_url() hasn't been overridden (defaults to /).
    """
    base_unauthorised_redirect_url = '/'

    def is_resource_protected(self, request, **kwargs):
        """
        Hook for determining if a resource should be protected.

        Returns True if and only if the middleware should be protecting
        the page or endpoint the request is trying to access.
        """
        return False

    def deny_access_condition(self, request, **kwargs):
        """
        Hook for disallowing access to a protected resource.

        Returns True if and only if the request should be disallowed
        from accessing the protected URL/page/endpoint.
        """
        return False

    def get_unauthorised_redirect_url(self, request):
        return self.base_unauthorised_redirect_url

    def get_access_denied_message(self, request):
        return ''

    def deny_access(self, request, **kwargs):
        """
        Standard failure behaviour.

        Returns HTTP 403 (Forbidden) for non-GET requests.

        For GET requests, returns HTTP 302 (Redirect) pointing at either a URL specified
        in the class's unauthorised_redirect attribute, if one exists, or / if not. This
        version also adds a (translated) message if one is passed in.
        """
        # Raise a 403 for POST/DELETE etc.
        if request.method != 'GET':
            return HttpResponseForbidden()

        # Add a message, if one has been defined.
        message = self.get_access_denied_message(request)
        if message:
            messages.info(request, _(message))

        # Return a HTTP 302 redirect.
        redirect_url = self.get_unauthorised_redirect_url(request)
        return redirect_to_login(request.path_info, login_url=redirect_url)

    def process_request(self, request):
        """
        The actual middleware method, called on all incoming requests.

        This default implementation will ignore the middleware (return None) if the
        conditions specified in is_resource_protected aren't met. If they are, it then
        tests to see if the user should be denied access via the denied_access_condition
        method, and calls deny_access (which implements failure behaviour) if so.
        """
        if not self.is_resource_protected(request):
            return

        if self.deny_access_condition(request):
            return self.deny_access(request)


class LoginPermissionMiddlewareMixin:
    """
    A mixin for middlewares related to user authentication.

    Provides implementations of deny_access_condition and get_access_denied_message that
    enforce that a user is authenticated.
    """
    def deny_access_condition(self, request, **kwargs):
        """Returns true if and only if the user isn't authenticated."""
        return request.user.is_anonymous()

    def get_access_denied_message(self, request):
        return 'You must be logged in to view this page.'


class UrlPermissionMiddleware(BasePermissionMiddleware):
    """
    Middleware that allows or denies access based on the resource's URL.

    The middleware is fed two lists of regular expressions intended to match URLs. One
    list defines URLs that are protected by this middleware, and the other defines URLs
    that are exempt from coverage. A URL that is both protected and not exempt will have
    its access controlled by this middleware.

    This class presents two hook methods - get_exempt_url_patterns (defaults to returning
    []) and get_protected_url_patterns (defaults to returning [r'^'], i.e. a regex that
    matches all URLs). Override these in order to supply your own URL lists to the
    middleware.
    """
    def get_exempt_url_patterns(self):
        """
        Hook method. Returns a list of regex patterns representing exempt URLs.

        The default implementation returns [] - no URLs are exempt.

        The regexes will need to be compiled. (The utils.compile_urls method helps here.)
        Override this method to supply your own list of exempt URLs (for instance, from
        Django settings) to this middleware.
        """
        return []

    def get_protected_url_patterns(self):
        """
        Hook method. Returns a list of regex patterns representing protected URLs.

        The default implementation returns [r'^'] - all URLs are protected.

        The regexes will need to be compiled. (The utils.compile_urls method helps here.)
        Override this method to supply your own list of protected URLs (for instance, from
        Django settings) to this middleware.
        """
        return ALL_URLS

    def is_resource_protected(self, request, **kwargs):
        """
        Returns true if and only if the resource's URL is *not* exempt and *is* protected.
        """
        exempt_urls = self.get_exempt_url_patterns()
        protected_urls = self.get_protected_url_patterns()
        path = request.path_info.lstrip('/')

        path_is_exempt = any(m.match(path) for m in exempt_urls)
        if path_is_exempt:
            return False

        path_is_protected = any(m.match(path) for m in protected_urls)
        if path_is_protected:
            return True

        return False



from unittest import skipIf, skipUnless, TestCase

import django
from django.contrib.auth import views
from django.core.urlresolvers import resolve, reverse
from django.views.generic import RedirectView


class URLsMixin(object):
    """
    A TestCase Mixin with a check_url helper method for testing urls.
    Pirated with slight modifications from incuna_test_utils
    https://github.com/incuna/incuna-test-utils/blob/master/incuna_test_utils/testcases/urls.py
    """

    def check_url(self, view_method, expected_url, url_name,
                  url_args=None, url_kwargs=None):
        """
        Assert a view's url is correctly configured

        Check the url_name reverses to give a correctly formated expected_url.
        Check the expected_url resolves to the correct view.
        """

        reversed_url = reverse(url_name, args=url_args, kwargs=url_kwargs)
        self.assertEqual(reversed_url, expected_url)

        # Look for a method rather than a class here
        # (just because of what we're testing)
        resolved_view_method = resolve(expected_url).func
        self.assertEqual(resolved_view_method.__name__, view_method.__name__)


class TestURLs(URLsMixin, TestCase):

    def test_login(self):
        self.check_url(
            views.login,
            '/login/',
            'login',
        )

    def test_logout(self):
        self.check_url(
            views.logout,
            '/logout/',
            'logout',
        )

    def test_password_change(self):
        self.check_url(
            views.password_change,
            '/password/change/',
            'password_change',
        )

    def test_password_change_done(self):
        self.check_url(
            views.password_change_done,
            '/password/change/done/',
            'password_change_done',
        )

    def test_password_reset(self):
        self.check_url(
            views.password_reset,
            '/password/reset/',
            'password_reset',
        )

    def test_password_reset_done(self):
        self.check_url(
            views.password_reset_done,
            '/password/reset/done/',
            'password_reset_done',
        )

    def test_password_reset_complete(self):
        self.check_url(
            views.password_reset_complete,
            '/password/reset/complete/',
            'password_reset_complete',
        )

    def test_sso(self):
        self.check_url(
            RedirectView.as_view(),
            '/sso/',
            'sso_login',
        )

    @skipIf(django.VERSION < (1, 6), 'This test is for a different Django version')
    def test_password_reset_confirm_uidb64(self):
        uidb64 = '09_AZ-az'
        token = '09AZaz-09AZaz'
        self.check_url(
            views.password_reset_confirm,
            '/password/reset/confirm/{0}/{1}/'.format(uidb64, token),
            'password_reset_confirm',
            url_kwargs={'uidb64': uidb64,
                        'token': token, },
        )

    @skipUnless(
        (1, 6) >= django.VERSION > (1, 7),
        'This test is for a different Django version',
    )
    def test_password_reset_confirm_uidb36(self):
        uidb36 = '09AZaz'
        token = '09AZaz-09AZaz'
        self.check_url(
            views.password_reset_confirm_uidb36,
            '/password/reset/confirm/{0}-{1}/'.format(uidb36, token),
            'password_reset_confirm',
            url_kwargs={'uidb36': uidb36,
                        'token': token, },
        )

    @skipIf(django.VERSION >= (1, 6), 'This test is for a different Django version')
    def test_password_reset_confirm_dj1_5(self):
        uidb36 = '09AZaz'
        token = '09AZaz-09AZaz'
        self.check_url(
            views.password_reset_confirm,
            '/password/reset/confirm/{0}-{1}/'.format(uidb36, token),
            'password_reset_confirm',
            url_kwargs={'uidb36': uidb36,
                        'token': token, },
        )



import mock

from django.db import models as django_models
from django.test import TestCase

from incuna_auth import models


CUSTOM_STATE = ('custom', 'Custom state')
INHERIT_STATE = (
    'base_inherit',
    'Inherit from parent (allow all users if no parent exists)',
)
BASE_STATES = (
    ('base_all', 'All users'),
    ('base_auth', 'Authenticated users only'),
    INHERIT_STATE,
)


class TestAccessStateExtensionMixin(TestCase):
    """Test that AccessStateExtensionMixin does what it's supposed to do."""

    class AccessState(models.AccessStateExtensionMixin):
        CUSTOM_STATES = (CUSTOM_STATE,)

    def test_custom_states(self):
        """Assert that setting the CUSTOM_STATE attribute works properly."""
        expected_states = (CUSTOM_STATE,) + BASE_STATES
        self.assertEqual(expected_states, self.AccessState.ACCESS_STATES)

    def test_handle_model(self):
        """
        Assert that add_to_class is called twice on the model, with correct parameters.

        This is done by using mock.MagicMock() as a model and handing that to
        the AccessState instance we're using. We can then assert how add_to_class was
        called without needing to introspect the model.
        """
        access = self.AccessState()
        model = mock.MagicMock()
        access.model = model

        access.handle_model()

        # We expect add_to_class to be called exactly twice in handle_model().
        self.assertEqual(model.add_to_class.call_count, 2)

        expected_states = (CUSTOM_STATE,) + BASE_STATES
        expected_field = django_models.CharField(
            max_length=255,
            choices=expected_states,
            default=INHERIT_STATE,
        )

        # add_to_class should be called with the tuple of states first,
        # then with the CharField.
        model.add_to_class.assert_calls([
            mock.call('ACCESS_STATES', expected_states),
            mock.call('access_state', expected_field),
        ])

    def test_handle_modeladmin(self):
        """
        Assert that add_extension_options is called once on the modeladmin.

        This is also done with mocking (see above). For this test, we can leave the
        model attribute on the AccessState instance as the default None.
        """
        access = self.AccessState()
        modeladmin = mock.MagicMock()

        access.handle_modeladmin(modeladmin)

        modeladmin.add_extension_options.assert_called_once_with('access_state')
